import Parser from 'tree-sitter';
import Bash from 'tree-sitter-bash';
import { debugLogObject } from '../util/debug.js';
import { stripQuotes } from '../util/string.js';
import * as FallbackResolver from './resolvers/fallback.js';
import * as KnownResolvers from './resolvers/index.js';
import { stripBinaryPath } from './util.js';
const parser = new Parser();
parser.setLanguage(Bash);
const getCommandsFromScript = (script) => {
    const tree = parser.parse(script);
    const commands = [];
    const traverse = (node) => {
        switch (node.type) {
            case 'command': {
                const commandNameIndex = node.children.findIndex(node => node.type === 'command_name');
                const command = node.children
                    .slice(commandNameIndex)
                    .map(node => node.text)
                    .map(stripQuotes);
                commands.push(command);
                break;
            }
            default:
                break;
        }
        for (const child of node.children) {
            traverse(child);
        }
    };
    traverse(tree.rootNode);
    return commands;
};
export const getBinariesFromScript = (script, { cwd, manifest, knownGlobalsOnly = false }) => {
    if (!script)
        return [];
    const fromArgs = (args) => getBinariesFromScript(stripQuotes(args.filter(arg => arg !== '--').join(' ')), { cwd, manifest });
    const commands = getCommandsFromScript(script);
    const getBinariesFromCommand = (command) => {
        const [bin, ...args] = command;
        const binary = stripBinaryPath(bin);
        if (!binary || binary === '.' || binary === 'source')
            return [];
        if (binary.startsWith('-') || binary.startsWith('"') || binary.startsWith('..'))
            return [];
        if (['bun', 'deno'].includes(binary))
            return [];
        if (binary in KnownResolvers) {
            return KnownResolvers[binary].resolve(binary, args, { cwd, manifest, fromArgs });
        }
        if (knownGlobalsOnly)
            return [];
        return FallbackResolver.resolve(binary, args, { cwd, manifest, fromArgs });
    };
    try {
        return commands.map(getBinariesFromCommand).flat();
    }
    catch (error) {
        debugLogObject('Bash parser error', error);
        return [];
    }
};
