import { compact } from '../../util/array.js';
import { getPackageNameFromModuleSpecifier } from '../../util/modules.js';
import { isInternal, dirname, toAbsolute } from '../../util/path.js';
import { load } from '../../util/plugin.js';
import { _resolve } from '../../util/require.js';
import { fallback } from './fallback.js';
const getDependencies = (config) => {
    const extendsSpecifiers = config.extends ? [config.extends].flat().map(resolveExtendsSpecifier) : [];
    if (extendsSpecifiers.includes('eslint-plugin-prettier'))
        extendsSpecifiers.push('eslint-config-prettier');
    const plugins = config.plugins ? config.plugins.map(resolvePluginPackageName) : [];
    const parser = config.parser;
    const extraPlugins = config.parserOptions?.babelOptions?.plugins ?? [];
    const extraParsers = config.parserOptions?.babelOptions?.presets ?? [];
    const settings = config.settings ? getDependenciesFromSettings(config.settings) : [];
    const overrides = config.overrides ? [config.overrides].flat().flatMap(getDependencies) : [];
    return compact([
        ...extendsSpecifiers,
        ...plugins,
        ...extraPlugins,
        parser,
        ...extraParsers,
        ...settings,
        ...overrides,
    ]);
};
export const getDependenciesDeep = async (configFilePath, dependencies = new Set(), options) => {
    const addAll = (deps) => deps.forEach(dependency => dependencies.add(dependency));
    const config = configFilePath.endsWith('package.json')
        ? options.manifest.eslintConfig
        : /(\.(jsonc?|ya?ml)|rc)$/.test(configFilePath)
            ? await load(configFilePath)
            : await fallback(configFilePath);
    if (config) {
        if (config.extends) {
            for (const extend of [config.extends].flat()) {
                if (isInternal(extend)) {
                    const filePath = toAbsolute(extend, dirname(configFilePath));
                    const extendConfigFilePath = _resolve(filePath);
                    dependencies.add(extendConfigFilePath);
                    addAll(await getDependenciesDeep(extendConfigFilePath, dependencies, options));
                }
            }
        }
        addAll(getDependencies(config));
    }
    return dependencies;
};
const resolvePackageName = (namespace, pluginName) => {
    return pluginName.includes(namespace + '-') || pluginName.endsWith(`/${namespace}`)
        ? pluginName
        : pluginName.startsWith('@')
            ? pluginName.includes('/')
                ? pluginName.replace(/\//, `/${namespace}-`)
                : `${pluginName}/${namespace}`
            : `${namespace}-${pluginName}`;
};
const resolvePluginPackageName = (pluginName) => resolvePackageName('eslint-plugin', pluginName);
const resolveExtendsSpecifier = (specifier) => {
    if (isInternal(specifier))
        return;
    if (/\/eslint-(config|plugin)/.test(specifier))
        return specifier;
    const strippedSpecifier = specifier
        .replace(/(^plugin:|:.+$)/, '')
        .replace(/\/(eslint-)?(recommended.*|stylistic.*|strict.*|all)$/, '');
    if (/eslint-(config|plugin)-/.test(strippedSpecifier))
        return strippedSpecifier;
    const packageName = getPackageNameFromModuleSpecifier(strippedSpecifier);
    if (!packageName)
        return;
    if (packageName === 'eslint')
        return;
    return resolvePackageName(specifier.startsWith('plugin:') ? 'eslint-plugin' : 'eslint-config', packageName);
};
const getDependenciesFromSettings = (settings = {}) => {
    return compact(Object.entries(settings).flatMap(([settingKey, settings]) => {
        if (settingKey === 'import/resolver') {
            return (typeof settings === 'string' ? [settings] : Object.keys(settings))
                .filter(key => key !== 'node')
                .map(key => `eslint-import-resolver-${key}`);
        }
        if (settingKey === 'import/parsers') {
            return typeof settings === 'string' ? [settings] : Object.keys(settings);
        }
    }));
};
